(function() {
  var Model, auth, createClient, crypto, end, forEach, hash, hashType, redis, redisModule, salt, util;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __indexOf = Array.prototype.indexOf || function(item) {
    for (var i = 0, l = this.length; i < l; i++) {
      if (this[i] === item) return i;
    }
    return -1;
  };
  crypto = require('crypto');
  redisModule = require('redis');
  util = require('util');
  salt = "";
  hashType = 'sha1';
  forEach = require('async').forEach;
  redis = void 0;
  hash = function(text) {
    return ((crypto.createHash(hashType)).update(text + salt)).digest('hex');
  };
  createClient = __bind(function(port, host, options) {
    module.exports.redis = redisModule.createClient(port, host, options);
    return redis = module.exports.redis;
  }, this);
  auth = function(password, callback) {
    return redis.auth(password, callback);
  };
  end = __bind(function() {
    return redis.end();
  }, this);
  Model = (function() {
    /*
        constructor
    
        @modelName: The name that will be used as the first token of all keys
        @fields: The main fields that will be used for storage.
        @indexFields: You must include any fields you will be using to perform lookups.
        */    var parent;
    function Model(modelName, fields, indexFields) {
      this.modelName = modelName;
      this.fields = fields;
      this.indexFields = indexFields != null ? indexFields : [];
      redis.get("" + this.modelName + ":id", function(err, data) {
        if (err != null) {
          return redis.set("" + this.modelName + ":id", 1);
        }
      });
    }
    parent = Model;
    /*
        delete method
        
        id: The id of the instance you are going to destroy.
        callback: Called with either the error returned by Redis or a true value
                  for a successful deletion.
        */
    Model.prototype["delete"] = function(id, callback) {
      var i, prefix, rcall;
      i = 0;
      prefix = "" + this.modelName + ":" + id + ":";
      rcall = __bind(function(field, nextField) {
        var passBack;
        passBack = function() {
          return redis.del(prefix + field, function(err) {
            if (err != null) {
              throw err;
            } else {
              return nextField();
            }
          });
        };
        if (__indexOf.call(this.indexFields, field) >= 0) {
          return this.reIndex(field, function(res) {
            if ((res != null) && res) {
              return passBack();
            } else {
              throw "Reindexing failed.";
            }
          });
        } else {
          return passBack();
        }
      }, this);
      return forEach(this.fields, rcall, function(err) {
        if (err != null) {
          throw err;
        } else {
          return callback(true);
        }
      });
    };
    /*
        load method
        
        id: The id of the instance you are loading.
        callback: Called with either an instance of the object you are loading
                  or the error returned by Redis.
        */
    Model.prototype.load = function(id, callback) {
      var i, prefix, rcall, robj;
      i = -1;
      prefix = "" + this.modelName + ":" + id + ":";
      robj = {
        "id": id
      };
      rcall = __bind(function(err, data) {
        var curfield;
        i++;
        if (err != null) {
          throw err;
          return callback(err);
        } else {
          if (i < this.fields.length) {
            curfield = this.fields[i];
            robj[curfield] = data;
            return redis.get(prefix + this.fields[i + 1], rcall);
          } else {
            this.attachSave(robj);
            return callback(robj);
          }
        }
      }, this);
      return redis.get(prefix + this.fields[0], rcall);
    };
    /*
        findAllBy method
        Returns all object instances with a specific value in a specific field.
        
        field: Name of the field being queried.
        value: The specific value to find.
        callback: Called either with list of instances or a null response.
        */
    Model.prototype.findAllBy = function(field, value, callback) {
      if (this.indexFields.indexOf(field > -1)) {
        return redis.lrange("" + this.modelName + ":" + field + ":" + value, 0, -1, __bind(function(err, data) {
          var i, icb, instances;
          if (err != null) {
            throw err;
            callback(null);
          }
          if ((err != null) || !(data != null)) {
            callback(null);
            return;
          }
          instances = [];
          i = 0;
          icb = __bind(function(instance) {
            if (instance != null) {
              instances.push(instance);
              i++;
            }
            if (i < data.length) {
              return this.load(data[i], icb);
            } else {
              return callback(instances);
            }
          }, this);
          return icb();
        }, this));
      }
    };
    /*
        findBy method
        Returns the first object instance  with a specific value in a specific field.
        Intended for unique values, but not enforced.
    
        field: Name of the field being queried.
        value: The specific value to find.
        callback: Called eitehr with the instance or a null response.
        */
    Model.prototype.findBy = function(field, value, callback) {
      if (this.indexFields.indexOf(field > -1)) {
        return redis.lrange("" + this.modelName + ":" + field + ":" + value, 0, 0, __bind(function(err, data) {
          if (err != null) {
            throw err;
            callback(err);
          }
          if (!(data != null)) {
            return callback(null);
          } else {
            return this.load(data[0], function(instance) {
              return callback(instance);
            });
          }
        }, this));
      } else {
        return callback(null);
      }
    };
    /*
        loadAll method
        Loads every instance of this model.
    
        callback: Is called with either the list of instances or the error
                  Redis returns.
        */
    Model.prototype.loadAll = function(callback) {
      var i, maxI, prefix, rcall, robjs;
      i = 1;
      maxI = -1;
      prefix = "" + this.modelName + ":id";
      robjs = [];
      rcall = __bind(function(data) {
        if (data != null) {
          robjs.push(data);
        }
        i++;
        if (i < maxI) {
          return this.load(i, rcall);
        } else {
          return callback(robjs);
        }
      }, this);
      return redis.get(prefix, __bind(function(err, data) {
        if (!(err != null) && (data != null)) {
          maxI = data;
          return this.load(1, rcall);
        } else {
          return callback(err);
        }
      }, this));
    };
    /*
        attachSave method
        Used by parts of the model that build instances to add a save method.
        This should not be used directly.
        instance: guess
        */
    Model.prototype.attachSave = function(instance) {
      var save;
      save = __bind(function(callback) {
        var numFields, prefix, rcall;
        prefix = "" + this.modelName + ":" + instance.id + ":";
        numFields = this.fields.length;
        rcall = __bind(function(field, nextField) {
          return redis.set(prefix + field, instance[field], __bind(function(err) {
            if (err != null) {
              throw err;
            } else {
              if (__indexOf.call(this.indexFields, field) >= 0) {
                return this.reIndex(field, function(res) {
                  if ((res != null) && res) {
                    return nextField();
                  } else {
                    throw "Reindex failed";
                  }
                });
              } else {
                return nextField();
              }
            }
          }, this));
        }, this);
        return forEach(this.fields, rcall, function(err) {
          if (err != null) {
            throw err;
          } else {
            return callback(instance);
          }
        });
      }, this);
      return instance.save = save;
    };
    /*
        create method
    
        Creates a new instance and saves it in Redis right away.
        data: The data, minus the id that goes into the model's fields.
        callback: Is called either with either the new instance or the error
                  Redis returns.
        */
    Model.prototype.create = function(data, callback) {
      return redis.incr("" + this.modelName + ":id", __bind(function(err, id) {
        if (!(err != null)) {
          data.id = id;
          this.attachSave(data);
          return data.save(__bind(function(res) {
            return callback(res);
          }, this));
        } else {
          throw err;
        }
      }, this));
    };
    /*
        reIndex method
    
        Rewrites an entire index from scratch. 
        */
    Model.prototype.reIndex = function(field, callback) {
      var indexKey;
      indexKey = "" + this.modelName + ":" + field;
      return redis.keys("" + indexKey + ":*", __bind(function(err, iKeys) {
        var deleter, passBack;
        if (err != null) {
          throw err;
        } else {
          deleter = __bind(function(key, fcallback2) {
            return redis.del(key, __bind(function(err) {
              if (err != null) {
                throw err;
              }
              return fcallback2();
            }, this));
          }, this);
          passBack = __bind(function() {
            return redis.keys("" + this.modelName + ":*:" + field, __bind(function(err, keys) {
              var indexer;
              indexer = __bind(function(key, lcallback) {
                var id, tokens;
                tokens = key.split(':');
                id = tokens[1];
                return redis.get(key, __bind(function(err, data) {
                  if (!(err != null) && data) {
                    return redis.lpush("" + indexKey + ":" + data, id, __bind(function(err) {
                      if (err != null) {
                        throw err;
                      }
                      return lcallback();
                    }, this));
                  } else if (err != null) {
                    throw err;
                    return lcallback();
                  }
                }, this));
              }, this);
              return forEach(keys, indexer, __bind(function(err) {
                if (err != null) {
                  throw err;
                  return callback(false);
                } else {
                  return callback(true);
                }
              }, this));
            }, this));
          }, this);
          if (iKeys.length > 0) {
            return forEach(iKeys, deleter, __bind(function(err) {
              if (err != null) {
                throw err;
              }
              return passBack();
            }, this));
          } else {
            return passBack();
          }
        }
      }, this));
    };
    /*
        reIndexAll method
    
        Rewrites all indexes from scratch.
        */
    Model.prototype.reIndexAll = function(callback) {
      var indexer;
      indexer = __bind(function(field, fcallback) {
        return this.reIndex(field, __bind(function(res) {
          return fcallback();
        }, this));
      }, this);
      return forEach(this.indexFields, indexer, function(err) {
        if (err) {
          throw err;
        } else {
          return callback(true);
        }
      });
    };
    return Model;
  })();
  module.exports.hashType = hashType;
  module.exports.hash = hash;
  module.exports.salt = salt;
  module.exports.createClient = createClient;
  module.exports.end = end;
  module.exports.Model = Model;
  module.exports.redis = redis;
}).call(this);
